#include <stdio.h>
#include <input.h>

#include "envs.h"

typedef unsigned char BYTE;

// ZXpand needs to be configured with M=L
//
#pragma output hrgpage = 8192

BYTE* textcurs;	// text (etc) cursor (draw) pos
BYTE* envimgad;	// envelope image address

BYTE regs[14];

typedef struct
{
	BYTE x,y;
	unsigned int value;
	BYTE bitwidth;
}
dataItem;

dataItem dataItems[18] =
{
	{7,2*14,388,12},{19,2*14,10,4},{25,2*14,0,1},{28,2*14,1,1},{31,2*14,0,1},
	{7,3*14,194,12},{19,3*14,10,4},{25,3*14,0,1},{28,3*14,0,1},{31,3*14,0,1},
	{7,4*14,97,12},{19,4*14,10,4},{25,4*14,0,1},{28,4*14,0,1},{31,4*14,0,1},
	{11,6*14,16,5},
	{11,7*14,5000,16},
	{11,9*14,0,4}
};

// state transition table - up,down,left,right
//
BYTE landingPads[18*4] =
{
	17,5,4,1,   17,6,0,2,   12,7,1,3,   13,8,2,4,   14,9,3,0,
	0,10,9,6,   1,11,5,7,   2,12,6,8,   3,13,7,9,   4,14,8,5,
	5,15,14,11, 6,15,10,12, 7,2,11,13,  8,3,12,14,  9,4,13,10,
	10,16,10,16,
	15,17,15,17,
	16,0,16,0
};

BYTE env2imgMap[16] =
{
	0,0,0,0,
	1,1,1,1,
	2,3,4,5,
	6,7,8,9
};




void __FASTCALL__ out(int portval)
{
   #asm
   ld b,h
   rrc b
   rrc b
   rrc b
   ld c,7
   out (c),l
   #endasm
}


void clrhr()
{
   	#asm
   	ld	de,$2001
   	ld	hl,$2000
   	ld	bc,24*32*8
	xor	a
   	ld	(hl),a
   	ldir
   	#endasm
}


// ZXpand needs to have overlay disabled
// in order to be able to access the ROM bitmap data
//
void __FASTCALL__ hrchar(int c)
{
   	#asm

	add		hl,hl		; character *= 8
	add		hl,hl
	add		hl,hl

	ld		de,$1e00	; offset into rom
	add		hl,de

	ld		de,(_TEXTCURS) ; screen location

	ex		de,hl

	ld		bc,32		; offset to next line
	ld		a,8			; loop count

phrc_loop:
	push	af

	ld		a,(de)		; copy byte to screen
	ld		(hl),a
	inc		de			; advance character data pointer
	add		hl,bc		; advance scan line

	pop		af
	dec		a
	jr		nz,phrc_loop

	#endasm
}


// draw a line of length L characters in colour H
//
void __FASTCALL__ hrline(int colourAndLength)
{
   	#asm
	ld		de,(_TEXTCURS)
	ld		a,h
hrl_loop:
	ld		(de),a
	inc		de
	dec		l
	jr		nz,hrl_loop
	#endasm
}


void putstringtohr(int x, int y, char* string)
{
 	textcurs = 8192 + x + (32 * y);
	while (*string)
	{
		hrchar(ascii_zx(*string));
		++textcurs;
		++string;
	}
}


void putchartohr(int x, int y, char c)
{
 	textcurs = 8192 + x + (32 * y);
	hrchar(ascii_zx(c));
}


// todo - add leading zero suppression?
//
void putnumtohrX(int x, int y, unsigned int num, BYTE digits)
{
	unsigned int n;
 	textcurs = 8192 + x + (32 * y);

	switch (digits)
	{
		case 5:
			n = num / 10000;
			hrchar(28 + n);
			num -= n * 10000;
			++textcurs;

		case 4:
			n = num / 1000;
			hrchar(28 + n);
			num -= n * 1000;
			++textcurs;

		case 3:
			n = num / 100;
			hrchar(28 + n);
			num -= n * 100;
			++textcurs;

		case 2:
			n = num / 10;
			hrchar(28 + n);
			num -= n * 10;
			++textcurs;
			hrchar(28 + num);
	}
}


void puthextohr(int x, int y, BYTE num)
{
 	textcurs = 8192 + x + (32 * y);

	hrchar(28 + (num >> 4));
	++textcurs;
	hrchar(28 + (num & 0x0f));
}


void draw(BYTE cell)
{
	if (dataItems[cell].bitwidth == 16)
	{
		putnumtohrX(dataItems[cell].x, dataItems[cell].y, dataItems[cell].value, 5);
	}
	else if(dataItems[cell].bitwidth == 12)
	{
		putnumtohrX(dataItems[cell].x, dataItems[cell].y, dataItems[cell].value, 4);
	}
	else if(dataItems[cell].bitwidth > 1)
	{
		// 8, 5, 4
		putnumtohrX(dataItems[cell].x, dataItems[cell].y, dataItems[cell].value, 2);
	}
	else
	{
		putchartohr(dataItems[cell].x, dataItems[cell].y, dataItems[cell].value ? '+' : '-');
	}

	if (cell == 17)
	{
		drawEnvelope();
	}
}


void underline(BYTE cell, BYTE black, BYTE two)
{
	int drawval = 0;
	int wid = (int)dataItems[cell].bitwidth;

	// careful here!
	//
	if (wid == 4 || wid == 5) wid = 2;
	if (wid == 12) wid = 4;
	if (wid == 16) wid = 5;

	textcurs = (BYTE*)(8192 + dataItems[cell].x + (32 * (dataItems[cell].y + 9)));

	if (black)
	{
		drawval = 0xff00;
	}

	drawval += wid;

	hrline(drawval);
	textcurs += 32;
	hrline(drawval);

	if (two)
	{
		textcurs -= 384;
		hrline(drawval);
		textcurs += 32;
		hrline(drawval);
	}
}


BYTE transitionState(BYTE curstate, BYTE move)
{
	BYTE i = landingPads[(curstate * 4) + move];

	underline(curstate, 0, 0);
	underline(i, 1, 0);

	in_WaitForNoKey();

	return i;
}


void makeSomeNoise(void)
{
	#asm
	ld		hl,_REGS
	ld		bc,$0e00

mns_allregs:
	ld		a,c
	out		($cf),a
	ld		a,(hl)
	out		($0f),a
	inc		c
	inc		hl
	djnz	mns_allregs

	#endasm
}


void shutUp(void)
{
	#asm
	ld		a,8
	ld		bc,$031f
	ld		e,00

stfu_loop:
	out		($cf),a
	out		(c),e
	inc		a
	djnz	stfu_loop
	#endasm
}


void assembleAYregs(void)
{
	regs[0] = dataItems[0].value & 0xff;		// chan a pitch fine
	regs[1] = dataItems[0].value / 256;			// chan a pitch coarse

	regs[2] = dataItems[5].value & 0xff;		// chan b pitch fine
	regs[3] = dataItems[5].value / 256;			// chan b pitch coarse

	regs[4] = dataItems[10].value & 0xff;		// chan c pitch fine
	regs[5] = dataItems[10].value / 256;		// chan c pitch coarse

	regs[6] = dataItems[15].value;				// noise frequency

	regs[7] = 0xc0 |							// mixer - bits are active low, set io as input [00------]
				dataItems[2].value << 3 |		// a noise
				dataItems[3].value |			// a tone
				dataItems[7].value << 4 |		// b noise
				dataItems[8].value << 1 |		// b tone
				dataItems[12].value << 5 |		// c noise
				dataItems[13].value << 2;		// c tone
	regs[7] ^= 0xff;

	regs[8]  = dataItems[4].value  ? 16 : dataItems[1].value;		// chan a vol/env control
	regs[9]  = dataItems[9].value  ? 16 : dataItems[6].value;		// chan b vol/env control
	regs[10] = dataItems[14].value ? 16 : dataItems[11].value;		// chan b vol/env control

	regs[11] = dataItems[16].value & 0xff;		// envelope frequency
	regs[12] = dataItems[16].value / 256;

	regs[13] = dataItems[17].value;				// envelope shape
}


void drawEnvelope(void)
{
	// each envelope image is 96px x 10px
	//
	envimgad = &envelopeData[(120 * env2imgMap[dataItems[17].value])];
	textcurs = (BYTE *)(8192 + 15 + (9*14 * 32) - 32);

	#asm
	ld		hl,(_ENVIMGAD)
	ld		de,(_TEXTCURS)
	ld		b,10

dre_loop:
	push	bc

	ld		bc,12
	ldir

	ld		a,e
	and		a
	add		a,20
	ld		e,a
	ld		a,d
	adc		0
	ld		d,a

	pop		bc
	djnz	dre_loop

	#endasm
}


// todo - replace this with some kind of input state handler
//
void editVal(BYTE cell)
{
	unsigned int max = (1 << dataItems[cell].bitwidth) - 1;

	putstringtohr( 0, 20*8-3, "CURSORS - CHANGE   0 - MOVE MODE");

	underline(cell, 1, 1);

	in_WaitForNoKey();

	while(1)
	{
		int key = in_Inkey();
		int v17 = dataItems[17].value;

		if (key == '0')
		{
			break;
		}

		if (key == '6')
		{
			--dataItems[cell].value;
		}
		else if (key == '7')
		{
			++dataItems[cell].value;
		}
		else if (key == '8')
		{
			dataItems[cell].value += 15;
		}
		else if (key == '5')
		{
			dataItems[cell].value -= 15;
		}
		else if (key == 'P')
		{
			assembleAYregs();
			makeSomeNoise();
			in_WaitForNoKey();
		}
		else if (key == 'Q')
		{
			shutUp();
			in_WaitForNoKey();
		}

		dataItems[cell].value &= max;
		draw(cell);

		/// EEEEEEEW
		if (v17 != dataItems[17].value)
		{
			in_WaitForNoKey();
		}

	}

	putstringtohr( 0, 20*8-3, " CURSORS - MOVE   0 - EDIT MODE ");

	underline(cell, 0, 1);
	underline(cell, 1, 0);
}





main()
{
	int key;
	BYTE state;

	out(0x07b0); // disable ZXpand overlay in order to access the character set

	clrhr();	// todo - use built-in HR screen allocation to increase compatibility

	putstringtohr( 0, 0,    "AY SOUND DESIGNER V1.0       SMB");

	putstringtohr( 0, 20*8-3, " CURSORS - MOVE   0 - EDIT MODE ");
	putstringtohr( 0, 21*8, "  P - PLAY SOUND     Q - QUIET  ");
	putstringtohr( 0, 23*8, "      PRESS A KEY TO QUIT.");

	putstringtohr( 0, 2*14, "FREQ-A ....  VOL-A ..   N. T. E.");
	putstringtohr( 0, 3*14, "FREQ-B ....  VOL-B ..   N. T. E.");
	putstringtohr( 0, 4*14, "FREQ-C ....  VOL-C ..   N. T. E.");

	putstringtohr( 0, 6*14, "FREQ-NOISE ..");
	putstringtohr( 0, 7*14, "FREQ-ENV   .....");
	putstringtohr( 0, 9*14, "ENV-SHAPE  ..");

	for (state = 0; state < 18; ++state)
	{
		draw(state);
	}

	key = 0;
	state = 0;

	underline(state, 1, 0);

	while(key != 'A')
	{
		key = in_Inkey();
		if (key == '7')
		{
			state = transitionState(state, 0);
		}
		else if (key == '6')
		{
			// down
			state = transitionState(state, 1);
		}
		else if (key == '5')
		{
			// left
			state = transitionState(state, 2);
		}
		else if (key == '8')
		{
			// right
			state = transitionState(state, 3);
		}
		else if (key == '0')
		{
			if (dataItems[state].bitwidth == 1)
			{
				dataItems[state].value = 1 - dataItems[state].value;
				draw(state);
			}
			else
			{
				editVal (state);
			}
			in_WaitForNoKey();
		}
		else if (key == 'P')
		{
			assembleAYregs();
			makeSomeNoise();
			in_WaitForNoKey();
		}
		else if (key == 'Q')
		{
			shutUp();
			in_WaitForNoKey();
		}
	}

	out(0x07b1); // re-enable ZXpand overlay
}
